"""
Colors module.
"""

from colorama import Fore, Back, Style, init

# Initialize Colorama
init(autoreset=True)


# Colors for ********************* TEXT *********************
def green(text):
    return Fore.GREEN + text + Fore.RESET


def blue(text):
    return Fore.BLUE + text + Fore.RESET


def yellow(text):
    return Fore.YELLOW + text + Fore.RESET


def red(text):
    return Fore.RED + text + Fore.RESET

  
def dim(text):
    return Style.DIM + text + Style.RESET_ALL



# Colors for ********************* BackGround *********************
def success(text):
    return Fore.WHITE + Back.GREEN + text


def warning(text):
    return Fore.WHITE + Back.YELLOW + text


def danger(text):
    return Fore.WHITE + Back.RED + text



"""
CreatedOn module.
"""

from datetime import datetime


class CreatedOn:
    """
    Class representing the creation date of an object.

    Attributes:
        created_on (datetime): The datetime object representing when the instance was created.
    """

    def __init__(self) -> None:
        """
        Initializes a new CreatedOn instance with the current date and time.

        Returns:
            None
        """
        self.created_on = datetime.now()

    def __str__(self) -> str:
        """
        Returns a string representation of the creation date.

        Returns:
            str: The creation date formatted as DD.MM.YYYY.
        """
        return self.created_on.strftime("%d.%m.%Y")



from colorama import Fore, Style


def get_help(_) -> str:
    """
    Returns a help message with list of available commands.
    """
    message = f"""
    {Fore.CYAN}Available commands:
    - hello: Displays a welcome message.
    - help: Shows this help message.
    - add-contact: Starts the process of adding a new contact.
                        Will ask for name, phone(s), birthday, email, and address.
    - change-contact: Opens a submenu with commands to change a contact:
        - add-phones: Adds a new phone.
        - remove-phones: Removes phone.
        - add-birthday: Adds a birthday.
        - edit-birthday: Edits the birthday.
        - remove-birthday: Removes the birthday.
        - add-email: Adds an email.
        - edit-email: Edits the email.
        - remove-email: Removes the email.
        - add-address: Adds an address.
        - edit-address: Edits the address.
        - remove-address: Removes the address.
    - delete-contact: Deletes a contact by name.
    - all-contacts: Shows all contacts with their phone numbers.
    - birthdays: Shows upcoming birthdays within the next 7 days.
    - fake-contacts: Generates a specified number of fake contacts and adds them to an address book.
    - search-contacts: Allows you to search for contacts by name or phone number.
    - smart-search: Interactive search with autocomplete suggestions during input.
    - add-note: Starts the process of adding a new note to the notebook.
                        Will ask for text, tag(s), reminder.
    - change-note: Opens a submenu with commands to change the note: 
        - edit-text: Edits the text of the note.
        - add-tags: Adds a tag to the note.
        - remove-tag: Removes the tag from the note.
        - add-reminder:  Adds a reminder to the note.
    - delete-note: Deletes a note from the notebook.
    - all-notes: Shows all notes in the notebook.
    - fake-notes: Generates a specified number of fake notes and adds them to the notebook.
    - reminders: Shows reminders within the next specified number of days.

    - close / exit : Exits the program.{Style.RESET_ALL}
    """
    return message



from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit import PromptSession
from prompt_toolkit.styles import Style


class CustomCompleter(Completer):
    """
    Custom CLI completter
    """

    def __init__(self, commands: list, all_commands: bool = False) -> None:
        super().__init__()
        self.commands = commands
        self.all_commands = all_commands

    def get_completions(self, document, complete_event):
        for command in self.commands:
            if (
                command[: len(document.current_line)] != document.current_line
                and not self.all_commands
            ):
                continue
            yield Completion(
                command,
                start_position=-len(document.current_line),
                style="bg:green fg:ansiblack",
                selected_style="fg:lightcyan bg:ansiblack",
            )


class Prompt():
    """
    Custom prompt
    """
    def __init__(self, mouse_support=False) -> None:
        self.session = PromptSession()
        self.mouse_support = mouse_support

    def prompt(self,
               message: str,
               commands: list,
               all_commands: bool = False,
               style: str = '') -> str:
        """
        Provide CLI prompt and return entered data
        """
        color_style = Style.from_dict({'prompt': style})
        return self.session.prompt(
                    message=[('class:prompt', message)],
                    completer=CustomCompleter(commands, all_commands),
                    style=color_style,
                    mouse_support=self.mouse_support
                )



"""
Birthday field module.
"""

from datetime import datetime
from ..constants.validation import DATE_FORMAT, validation_errors


class Birthday:
    """
    Class representing a birthday field.
    """

    def __init__(self, value: str) -> None:
        try:
            birthday = datetime.strptime(value, DATE_FORMAT).date()
            if birthday > datetime.now().date():
                raise ValueError(validation_errors["future_birthday"])
            self.value = birthday
        except ValueError as exc:
            raise ValueError(validation_errors["invalid_birthday"]) from exc

    def __str__(self) -> str:
        return self.value.strftime(DATE_FORMAT)



"""
Controllers module.

The controllers module contains functions that interact with the user and
modify the notes book.
"""

from ..notes.notes_book import NotesBook
from ..notes.note import Note
from ..helpers.colors import green, blue, dim, success, warning, danger
from ..helpers.completer import Prompt
from ..helpers.generate_data import generate_random_note
from ..helpers.display import display_notes, display_note, display_reminders
from ..constants.questions import questions
from ..constants.info_messages import info_messages
from ..constants.commands import commands
from ..constants.validation import validation_errors


def add_note(book: NotesBook) -> str:
    """
    Adds a new note to the `book'.

    Args:
        book (NotesBook): An instance of the `NotesBook` class.

    Returns:
        str: A message indicating whether the note was added or
        if the input is invalid.
    """
    try:
        while True:
            title = input(dim(questions["back"]) + blue(questions["title"]))
            if title.strip().lower() in book.data:
                print(dim(questions["back"]) + warning(
                    validation_errors["duplicate_title"]
                ).format(title))
                continue
            try:
                new_note = Note(title)
                break
            except ValueError as e:
                print(dim(questions["back"]) + danger(str(e)))
                continue

        edit_text(new_note)
        add_tags(new_note)
        edit_reminder(new_note)
        book.add_note(new_note)
        print(display_note(new_note))
        return success(info_messages["note_added"])
    except KeyboardInterrupt:
        return danger("\n" + info_messages["operation_cancelled"])


def change_note(book: NotesBook) -> str:
    """
    Edits an existing note in the `notes book`.

    Args:
        book (NotesBook): An instance of the `NotesBook` class.

    Returns:
        str: A message indicating whether the note was edited or not.
    """
    is_edited = False
    prompt = Prompt()
    try:
        while True:
            question = dim(questions["back"]) + blue(questions["title"])
            print(question, end="")
            title = prompt.prompt(
                " " * len(questions["back"] + questions["title"]), list(book)
            )
            print("\033[F\033[K", end="")
            print(f"{question}{title}")
            try:
                note = book.find(title)
                break
            except ValueError as e:
                print(dim(questions["back"]) + danger(str(e)))
                continue

        print(display_note(note))
        while True:
            all_commands = {
                commands["main_menu"]: None,
                commands["add_tags"]: add_tags,
            }
            if note.add_tags:
                all_commands[commands["remove_tag"]] = remove_tag
            if note.text:
                all_commands[commands["edit_text"]] = edit_text
                all_commands["remove_text"] = remove_text
            else:
                all_commands["add_text"] = edit_text
            if note.reminder:
                all_commands[commands["remove_reminder"]] = remove_reminder
                all_commands[commands["edit_reminder"]] = edit_reminder
            else:
                all_commands[commands["add_reminder"]] = edit_reminder
            options = f"Options: {", ".join(list(all_commands))}"
            print(dim(questions["back"] + options))
            question = dim(questions["back"]) + blue(questions["command"])
            print(question, end="")
            n = len(questions["back"] + questions["command"])
            command = prompt.prompt(" " * n, list(all_commands))
            print("\033[F\033[K", end="")
            print(f"{question}{command}")

            if command == commands["main_menu"]:
                break
            if command in all_commands:
                all_commands[command](note)
                is_edited = True
                print(display_note(note))
            else:
                print(
                    dim(questions["back"])
                    + danger(info_messages["unknown_command"])
                )
                continue
        return (
            success("\n" + info_messages["note_edited"])
            if is_edited
            else danger(info_messages["operation_cancelled"])
        )
    except KeyboardInterrupt:
        return (
            success("\n" + info_messages["note_edited"])
            if is_edited
            else danger(info_messages["operation_cancelled"])
        )


def edit_text(note: Note) -> None:
    """
    Edits text in the `note`.

    Args:
        note (Note): An instance of the `Note` class.

    Returns:
        None
    """
    while True:
        try:
            text = input(
                dim(questions["back"])
                + blue(questions["text"] + questions["skip"])
            )
            if not text:
                break
            note.add_text(text)
            print(
                dim(questions["back"])
                + green(info_messages["text_added"])
            )
            break
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def remove_text(note: Note) -> None:
    """
    Removes text from the `note`.

    Args:
        note (Note): An instance of the `Note` class.

    Returns:
        None
    """
    note.remove_text()
    print(dim(questions["back"]) + green(info_messages["text_removed"]))


def add_tags(note: Note) -> None:
    """
    Adds tag to the `note`.

    Args:
        note (Note): An instance of the `Note` class.

    Returns:
        None
    """
    while True:
        try:
            tags = input(
                dim(questions["back"])
                + blue(questions["tags"] + questions["skip"])
            )
            if not tags.strip():
                break
            note.add_tags(tags)
            print(dim(questions["back"]) + green(info_messages["tags_added"]))
            break
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def remove_tag(note: Note) -> None:
    """
    Delete the tag of the `note`.

    Args:
        note (Note): An instance of the `Note` class.
        tag (str): An instance of the `str` class.

    Returns:
        None
    """
    prompt = Prompt()
    while True:
        try:
            question = dim(questions["back"]) + blue(
                questions["tag"] + questions["skip"]
            )

            print(question, end="")
            n = len(questions["back"] + questions["tag"] + questions["skip"])
            tag = prompt.prompt(" " * n, [tag.value for tag in note.tags])
            print("\033[F\033[K", end="")
            print(f"{question}{tag}")
            if not tag:
                break
            note.remove_tag(tag)
            print(
                dim(questions["back"])
                + green(info_messages["tag_removed"])
            )
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def edit_reminder(note: Note) -> None:
    """
    Edits reminder in the `note`.

    Args:
        note (Note): An instance of the `Note` class.

    Returns:
        None
    """
    while True:
        try:
            reminder = input(
                dim(questions["back"])
                + blue(questions["reminder"] + questions["skip"])
            )
            if not reminder:
                break
            note.set_reminder(reminder)
            print(
                dim(questions["back"])
                + green(info_messages["reminder_added"])
            )
            break
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def remove_reminder(note: Note) -> None:
    """
    Removes reminder from the `note`.

    Args:
        note (Note): An instance of the `Note` class.

    Returns:
        None
    """
    note.remove_reminder()
    print(dim(questions["back"]) + green(info_messages["reminder_removed"]))


def get_notes(book: NotesBook) -> str:
    """
    Returns a string containing the title, text, tags, created_on, reminder of
    all notes in the `book`.

    Args:
        book (NotesBook): An instance of the `NotesBook` class.

    Returns:
        str: A string containing title, text, tags, created_on, reminder of
        all notes in the `book`.
    """
    if not book.data:
        return warning(info_messages["no_notes"])

    return display_notes(book)


def delete_note(book: NotesBook) -> str:
    """
    Deletes a note from the notes book.

    Args:
        book (NotesBook): An instance of the NotesBook class from which
        the note will be deleted.

    Returns:
        str: A success message indicating the deletion of the contact.
    """
    prompt = Prompt()
    while True:
        try:
            question = dim(questions["back"]) + blue(questions["title"])
            print(question, end="")
            title = prompt.prompt(
                " " * len(questions["back"] + questions["title"]), list(book)
            )
            print("\033[F\033[K", end="")
            print(f"{question}{title}")
            book.delete(title)
            break
        except ValueError as error:
            print(dim(questions["back"]) + danger(str(error)))
        except KeyboardInterrupt:
            return danger(info_messages["operation_cancelled"])
    return success(info_messages["note_deleted"])


def fake_notes(book: NotesBook) -> str:
    """
    Generates a specified number of fake notes and adds them to the NotesBook.

    Args:
        book (NotesBook): An instance of the NotesBook class to which the
        fake notes will be added.

    Returns:
        str: A success message indicating the number of fake notes added.
    """
    while True:
        try:
            count = input(dim(questions["back"]) + blue(questions["notes"]))
            if not count.isdigit():
                print(
                    dim(questions["back"])
                    + warning(validation_errors["invalid_number"])
                )
                continue
            count = int(count)
            break
        except KeyboardInterrupt:
            return danger(info_messages["operation_cancelled"])

    for _ in range(count):
        note_data = generate_random_note()
        note = Note(note_data["title"])
        if note_data["text"]:
            note.add_text(note_data["text"])
        for tag in note_data["tags"]:
            note.add_tags(tag)
        if note_data["reminder"]:
            note.set_reminder(note_data["reminder"])
        try:
            book.add_note(note)
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))

    return success(info_messages["fake_notes_generated"])


def reminders(book: NotesBook) -> str:
    """
    Returns a string containing the titles and reminder dates of all notes
    in the `book` that have a reminder within the next specified number of
    days.

    Args:
        book (NotesBook): An instance of the `NotesBook` class.

    Returns:
        str: A string containing the titles and reminder dates of all notes
        in the `book` that have a reminder within the next specified number
        of days.
        If there are no reminders within the specified period, a warning
        message is returned.
    """
    while True:
        days = input(dim(questions["back"]) + blue(questions["days"]))
        if not days.isdigit():
            print(
                dim(questions["back"])
                + warning(validation_errors["invalid_number"])
            )
            continue
        notes = book.upcoming_reminders(int(days))
        if not notes:
            return warning(info_messages["no_reminders"])

        return display_reminders(book, int(days))


def search_notes(book: NotesBook) -> str:
    """
    Searches for notes in the `NotesBook` that match the given search term.
    The search term can be a note title or a tag. If searching by tag,
    results are sorted by the presence of the tag.

    Args:
        book (NotesBook): An instance of the `NotesBook` class.

    Returns:
        str: A string containing the notes that match the search term.
    """
    while True:
        search_term = input(
            "Enter search term (title or tag) or 'q' to quit: ").strip()
        if search_term.lower() == "q":
            return "Operation canceled."

        if not search_term:
            print("Please enter a valid search term or 'q' to exit.")
            continue

        results_by_title = book.search_notes_by_title(search_term)
        if results_by_title:
            sorted_results_by_title = book.sort_notes_by_title(
                results_by_title)
            return "\n".join([str(note) for note in sorted_results_by_title])

        results_by_tag = book.search_notes_by_tag(search_term)
        if results_by_tag:
            sorted_results_by_tag = book.sort_notes_by_tag(search_term)
            return "\n".join([str(note) for note in sorted_results_by_tag])

        print("No notes found matching the search term.")



"""
Contacts controllers module.

The contacts controllers module contains functions that interact with the user
and modify the address book.
"""

from ..contacts.address_book import AddressBook
from ..contacts.record import Record
from ..helpers.colors import green, blue, dim, success, warning, danger
from ..helpers.generate_data import generate_random_contact
from ..helpers.completer import Prompt
from ..constants.questions import questions
from ..constants.info_messages import info_messages
from ..constants.commands import commands
from ..constants.validation import validation_errors


def add_contact(book: AddressBook) -> str:
    """
    Adds a new contact to the `book` if the name is not already in the
    `book` or adds the phone number to an existing contact.

    Args:
        book (AddressBook): An instance of the `AddressBook` class.

    Returns:
        str: A message indicating whether the contact was added or updated, or
        if the input is invalid.
    """
    try:
        while True:
            name = input(dim(questions["back"]) + blue(questions["name"]))
            if name.strip().lower() in book.data:
                print(
                    dim(questions["back"])
                    + warning(validation_errors["duplicate_name"]).format(name)
                )
                continue
            try:
                new_record = Record(name)
                break
            except ValueError as e:
                print(dim(questions["back"]) + danger(str(e)))
                continue

        add_phones(new_record)
        edit_birthday(new_record)
        edit_email(new_record)
        edit_address(new_record)
        book.add_record(new_record)
        print(book.display_contacts([new_record]))
        return success(info_messages["contact_added"])
    except KeyboardInterrupt:
        return danger("\n" + info_messages["operation_cancelled"])


def change_contact(book: AddressBook) -> str:
    """
    Edits an existing contact in the `book`.

    Args:
        book (AddressBook): An instance of the `AddressBook` class.

    Returns:
        str: A message indicating whether the contact was edited or not.
    """
    is_edited = False
    prompt = Prompt()
    try:
        while True:
            question = dim(questions["back"]) + blue(questions["name"])
            print(question, end="")
            name = prompt.prompt(
                " " * len(questions["back"] + questions["name"]), list(book)
            )
            print("\033[F\033[K", end="")
            print(f"{question}{name}")
            try:
                contact = book.find(name)
                break
            except ValueError as e:
                print(dim(questions["back"]) + danger(str(e)))
                continue

        print(book.display_contacts([contact]))
        while True:
            all_commands = {
                commands["main_menu"]: None,
                commands["add_phones"]: add_phones,
            }
            if contact.phones:
                all_commands[commands["remove_phone"]] = remove_phone
            if contact.birthday:
                all_commands[commands["remove_birthday"]] = remove_birthday
                all_commands[commands["edit_birthday"]] = edit_birthday
            else:
                all_commands[commands["add_birthday"]] = edit_birthday
            if contact.email:
                all_commands[commands["remove_email"]] = remove_email
                all_commands[commands["edit_email"]] = edit_email
            else:
                all_commands[commands["add_email"]] = edit_email
            if contact.address:
                all_commands[commands["remove_address"]] = remove_address
                all_commands[commands["edit_address"]] = edit_address
            else:
                all_commands[commands["add_address"]] = edit_address
            options = f"Options: {", ".join(list(all_commands))}"
            print(dim(questions["back"] + options))
            question = dim(questions["back"]) + blue(questions["command"])
            print(question, end="")
            n = len(questions["back"] + questions["command"])
            command = prompt.prompt(" " * n, list(all_commands))
            print("\033[F\033[K", end="")
            print(f"{question}{command}")

            if command == commands["main_menu"]:
                break
            if command in all_commands:
                all_commands[command](contact)
                is_edited = True
                print(book.display_contacts([contact]))
            else:
                print(
                    dim(questions["back"])
                    + danger(info_messages["unknown_command"])
                )
                continue
        return (
            success(info_messages["contact_edited"])
            if is_edited
            else danger(info_messages["operation_cancelled"])
        )
    except KeyboardInterrupt:
        return (
            success(info_messages["contact_edited"])
            if is_edited
            else danger(info_messages["operation_cancelled"])
        )


def add_phones(contact: Record) -> None:
    """
    Adds a new phone number to the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    while True:
        try:
            phone = input(
                dim(questions["back"])
                + blue(questions["phone"] + questions["skip"])
            )
            if not phone.strip():
                break
            contact.add_phone(phone)
            print(dim(questions["back"]) + green(info_messages["phone_added"]))
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def remove_phone(contact: Record) -> None:
    """
    Removes a phone number from the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    prompt = Prompt()
    while True:
        try:
            question = dim(questions["back"]) + blue(
                questions["phone"] + questions["skip"]
            )

            print(question, end="")
            n = len(questions["back"] + questions["phone"] + questions["skip"])
            phone = prompt.prompt(" " * n, list(map(str, contact.phones)))
            print("\033[F\033[K", end="")
            print(f"{question}{phone}")
            if not phone:
                break
            contact.remove_phone(phone)
            print(
                dim(questions["back"])
                + green(info_messages["phone_removed"])
            )
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def edit_birthday(contact: Record) -> None:
    """
    Edits the birthday of the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    while True:
        try:
            birthday = input(
                dim(questions["back"])
                + blue(questions["birthday"] + questions["skip"])
            )
            if not birthday:
                break
            contact.add_birthday(birthday)
            print(
                dim(questions["back"])
                + green(info_messages["birthday_added"])
            )
            break
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def remove_birthday(contact: Record) -> None:
    """
    Removes the birthday from the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    contact.remove_birthday()
    print(dim(questions["back"]) + green(info_messages["birthday_removed"]))


def edit_address(contact: Record) -> None:
    """
    Edits the address of the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    while True:
        try:
            address = input(
                dim(questions["back"])
                + blue(questions["address"] + questions["skip"])
            )
            if not address:
                break
            contact.add_address(address)
            print(
                dim(questions["back"])
                + green(info_messages["address_added"])
            )
            break
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def remove_address(contact: Record) -> None:
    """
    Removes the address from the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    contact.remove_address()
    print(dim(questions["back"]) + green(info_messages["address_removed"]))


def edit_email(contact: Record) -> None:
    """
    Edits the email of the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    while True:
        try:
            email = input(
                dim(questions["back"])
                + blue(questions["email"] + questions["skip"])
            )
            if not email:
                break
            contact.add_email(email)
            print(dim(questions["back"]) + green(info_messages["email_added"]))
            break
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))
            continue


def remove_email(contact: Record) -> None:
    """
    Removes the email from the `contact`.

    Args:
        contact (Record): An instance of the `Record` class.

    Returns:
        None
    """
    contact.remove_email()
    print(dim(questions["back"]) + green(info_messages["email_removed"]))


def delete_contact(book: AddressBook) -> str:
    """
    Deletes a contact from the address book.

    Args:
        book (AddressBook): An instance of the AddressBook class from which
        the contact will be deleted.

    Returns:
        str: A success message indicating the deletion of the contact.
    """
    prompt = Prompt()
    while True:
        try:
            question = dim(questions["back"]) + blue(questions["name"])
            print(question, end="")
            name = prompt.prompt(
                " " * len(questions["back"] + questions["name"]), list(book)
            )
            print("\033[F\033[K", end="")
            print(f"{question}{name}")
            book.delete(name)
            break
        except ValueError as error:
            print(dim(questions["back"]) + danger(str(error)))
        except KeyboardInterrupt:
            return danger(info_messages["operation_cancelled"])
    return success(info_messages["contact_deleted"])


def get_contacts(book: AddressBook) -> str:
    """
    Returns a string containing the names and phone numbers of all contacts
    in the `book`.

    Args:
        book (AddressBook): An instance of the `AddressBook` class.

    Returns:
        str: A string containing the names and phone numbers of all contacts
        in the `book`.
    """
    if not book.data:
        return warning(info_messages["no_contacts"])
    return book.display_contacts(book.data.values())


# @input_error
def birthdays(book: AddressBook) -> str:
    """
    Returns a string containing the names and congratulation dates of all
    contacts in the `book` who have a birthday within the next 7 days.

    Args:
        book (AddressBook): An instance of the `AddressBook` class.

    Returns:
        str: A string containing the names and congratulation dates of all
        contacts in the `book` who have a birthday within the next 7 days.
    """
    while True:
        days = input(dim(questions["back"]) + blue(questions["days"]))
        if not days.isdigit():
            print(
                dim(questions["back"])
                + warning(validation_errors["invalid_number"])
            )
            continue
        return book.upcoming_birthdays(int(days))


def fake_contacts(book: AddressBook) -> str:
    """
    Generates a specified number of fake contacts and adds them to the address
    book.

    Args:
        book (AddressBook): An instance of the AddressBook class to which the
        fake contacts will be added.

    Returns:
        str: A success message indicating the number of fake contacts added.
    """
    while True:
        try:
            count = input(dim(questions["back"]) + blue(questions["contacts"]))
            if not count.isdigit():
                print(
                    dim(questions["back"])
                    + warning(validation_errors["invalid_number"])
                )
                continue
            count = int(count)
            break
        except KeyboardInterrupt:
            return danger(info_messages["operation_cancelled"])

    for _ in range(count):
        contact = generate_random_contact()
        record = Record(contact["name"])
        for phone in contact["phones"]:
            record.add_phone(phone)
        if contact["birthday"]:
            record.add_birthday(contact["birthday"])
        if contact["email"]:
            record.add_email(contact["email"])
        if contact["address"]:
            record.add_address(contact["address"])
        try:
            book.add_record(record)
        except ValueError as e:
            print(dim(questions["back"]) + danger(str(e)))

    return success(info_messages["fake_contacts_generated"])


def search_contacts(book: AddressBook) -> str:
    """
    Searches for contacts in the `book` that match the given search term.

    Args:
        book (AddressBook): An instance of the `AddressBook` class.

    Returns:
        str: A string containing the names and phone numbers of all contacts
        in the `book` that match the given search term.
    """
    while True:
        search_term = input(blue("Enter search term (name or phone number): "))
        if search_term.lower() == "q":
            return danger("Operation canceled.")
        if not search_term:
            print(warning("Please enter a valid search term or q to exit"))
            continue

        results = book.search(search_term)
        if results:
            return book.display_contacts(results, search_term)
        return warning("No contacts found matching the search term.")


def interactive_search_with_autocomplete(book: AddressBook) -> str:
    """
    Interactive search with autocomplete suggestions during input.
    """
    # Get all names from the address book for autocomplete suggestions
    contact_names = [record.name.value for record in book.data.values()]

    prompt = Prompt()

    while True:
        try:
            search_term = prompt.prompt(
                "Enter search term (name or phone number): ", contact_names
            ).strip()
            if search_term.lower() == "q":
                return danger("Operation canceled.")
            if not search_term:
                print(warning("Please enter a valid search term or q to exit"))
                continue

            results = book.smart_search(search_term)
            if results:
                # Here we make sure to pass the search_term to highlight it
                return book.display_contacts(results, search_term)

            print(warning("No contacts found matching the search term."))
        except KeyboardInterrupt:
            return danger("\nOperation canceled.")



import pyfiglet


def print_title(title: str, color_func, font_size: str = 'standart'):
    font_map = {
        'small': 'slant',
        'medium': 'standard',
        'large': 'big'
    }
    font_name = font_map.get(font_size, 'standard')
    ascii_title = pyfiglet.figlet_format(title, font=font_name)
    print(color_func(ascii_title))



import re
from datetime import datetime


def validate_phone(phone):
    """Validate that the phone number consists of exactly 10 digits."""
    return phone.isdigit() and len(phone) == 10


def validate_name(name):
    """Validate that the name contains only letters and spaces."""
    return re.match("^[A-Za-z\s]+$", name) is not None


def validate_email(email):
    """Validate that the email has a valid format."""
    return re.match(r"^[\w\.-]+@[\w\.-]+\.\w+$", email) is not None


def validate_birthday(birthday):
    """Validate that the birthday is in the format DD.MM.YYYY."""
    try:
        datetime.strptime(birthday, "%d.%m.%Y")
        return True
    except ValueError:
        return False
    
    
def validate_address(address):
    """Validate that the address is not empty and contains only valid characters."""
    return bool(address) and re.match("^[A-Za-z0-9\s,.-]+$", address) is not None


"""
Display module.

The display module contains functions that display data to the user.
"""

import re
import textwrap
from tabulate import tabulate
from ..notes.notes_book import NotesBook
from ..notes.note import Note
from ..helpers.colors import blue, green, yellow, warning


def wrap_text(text: str, width: int = 20) -> str:
    """
    Wraps the given text to a specified width using textwrap.wrap.

    Args:
        text (str): The text to wrap.
        width (int, optional): The maximum width of each line. Defaults to 20.

    Returns:
        str: The wrapped text with lines separated by newline characters.
    """
    return "\n".join(textwrap.wrap(str(text), width=width))


def display_table(headers: list, table: list) -> str:
    """
    Displays a table with the given headers and data.

    Args:
        headers (list): A list of column headers.
        table (list): A list of rows, where each row is a list of values.

    Returns:
        str: The formatted table as a string.
    """
    colored_headers = [blue(header) for header in headers]
    return tabulate(table, headers=colored_headers, tablefmt="grid")


def highlight_term(text: str, term: str, color_code: str = "\033[43m") -> str:
    """
    Highlights the term in the text by changing the background color.
    Args:
        text (str): The text to search in.
        term (str): The term to highlight.
        color_code (str, optional): The ANSI background color code to use for
        highlighting (default is yellow).
    Returns:
        str: The text with the term highlighted with a background color.
    """
    term = re.escape(term)
    highlighted_text = re.sub(
        f'({term})',
        f'{color_code}\\1\033[0m',
        text,
        flags=re.IGNORECASE
    )
    return highlighted_text


def display_notes(book: NotesBook) -> str:
    if not book.data:
        return "Notes book is empty."

    headers = ["Title", "Text", "Tags", "Created On", "Reminder"]
    colored_headers = [blue(header) for header in headers]
    table = []
    for note in book.data.values():
        tags_str = (", ".join([f"#{tag}" for tag in note.tags])
                    if note.tags else " - ")
        reminder_str = str(note.reminder) if note.reminder else " - "
        table.append([
            str(note.title),
            wrap_text(note.text),
            wrap_text(tags_str),
            str(note.created_on),
            reminder_str
        ])

    table_str = tabulate(table, headers=colored_headers, tablefmt="grid")
    return table_str


def display_note(note: Note) -> str:
    headers = ["Title", "Text", "Tags", "Created On", "Reminder"]
    colored_headers = [green(header) for header in headers]
    tags_str = (", ".join([f"#{tag}" for tag in note.tags])
                if note.tags else " - ")
    reminder_str = str(note.reminder) if note.reminder else " - "
    table = [[
        str(note.title),
        wrap_text(note.text),
        wrap_text(tags_str),
        str(note.created_on),
        reminder_str
    ]]

    table_str = tabulate(table, headers=colored_headers, tablefmt="grid")
    return table_str


def display_reminders(book: NotesBook, days: int) -> str:
    notes_with_reminders = book.upcoming_reminders(days)
    if not notes_with_reminders:
        return warning(f"No reminders in the next {days} days.")
    headers = ["Title", "Text", "Tags", "Created On", "Reminder"]
    colored_headers = [yellow(header) for header in headers]
    table = []
    for note in notes_with_reminders:
        tags_str = (", ".join([f"#{tag}" for tag in note.tags])
                    if note.tags else " - ")
        reminder_str = str(note.reminder) if note.reminder else " - "
        table.append([
            str(note.title),
            wrap_text(note.text),
            wrap_text(tags_str),
            str(note.created_on),
            reminder_str
        ])

    table_str = tabulate(table, headers=colored_headers, tablefmt="grid")
    return table_str



"""
Tag field module
"""


class Tag:
    """
    Class representing a tag field.
    """

    def __init__(self, value: str) -> None:
        """
        Initialize a tag field.

        Args:
            value (str): The tag of the note.

        Raises:
            ValueError: If the value is empty.
        """
        if not value.strip():
            raise ValueError("Tag cannot be empty")
        if len(value.strip()) < 2 or len(value.strip()) > 16:
            raise ValueError("Tag must be min 2 symbols and max 16 symbols")
        self.value = value

    def __str__(self) -> str:
        """
        Return the string representation of the tag field.

        Returns:
            str: The string representation of the tag field.
        """

        return f"#{self.value}"

    def __eq__(self, other) -> bool:
        """
        Checks if Tag instance equal to other instance or str.

        Args:
            other (Tag): The Tag instance to compare with.

        Returns:
            bool: True if the Tag instances are equal, False otherwise.
        """
        if isinstance(other, Tag):
            return self.value == other.value
        return self.value == other



questions = {
    "command": "Enter a command: ",
    "name": "Enter name: ",
    "phone": "Enter phone number: ",
    "birthday": "Enter birthday: ",
    "email": "Enter email: ",
    "address": "Enter address: ",
    "title": "Enter title: ",
    "text": "Enter text: ",
    "tags": "Enter tags separated by spaces: ",
    "tag": "Enter tag: ",
    "reminder": "Enter reminder: ",
    "days": "Enter number of days: ",
    "contacts": "Enter number of contacts: ",
    "notes": "Enter number of notes: ",
    "skip": "(press Enter to skip) ",
    "back": "[back to main menu CTRL + C] ",
}



"""
Email field module.
"""

import re
from ..constants.validation import EMAIL_PATTERN, validation_errors


class Email:
    """
    Class representing an email field.
    """

    def __init__(self, email: str) -> None:
        """
        Initialize an email field.

        Args:
            email (str): The email address.

        Raises:
            ValueError: If the email does not meet the validation criteria.
        """
        if not re.match(EMAIL_PATTERN, email):
            raise ValueError(validation_errors["invalid_email"])
        self.value = email

    def __str__(self) -> str:
        """
        Return the string representation of the address field.

        Returns:
            str: The string representation of the address field.
        """
        return self.value



"""
Editor module.
"""

import tkinter as tk
from tkinter import font
from typing import Optional


class Editor(tk.Frame):
    """
    A class used to create a text editor with a character limit and keyboard
    shortcuts.

    Attributes:
        char_limit (int): The maximum number of characters allowed in the text
        entry.
        master (tk.Tk): The root window or parent widget.
        result (str): The text content saved by the user.
    """

    def __init__(self,
                 master: Optional[tk.Tk] = None,
                 char_limit: int = 500,
                 width: int = 1000,
                 height: int = 600,
                 initial_text: Optional[str] = ""):
        """
        Initializes a new Editor instance.

        Args:
            master (Optional[tk.Tk]): The root window or parent widget. If not
            provided, a new Tk instance is created.
            char_limit (int): The maximum number of characters allowed in the
            text entry.
            width (int): The width of the window.
            height (int): The height of the window.
            initial_text (Optional[str]): The initial text to be displayed in
            the text entry widget.
        """
        if master is None:
            master = tk.Tk()
        super().__init__(master=master, bg="black")
        self.master: tk.Tk = master
        self.char_limit: int = char_limit
        self.result: str = ""
        self.initial_text: str = initial_text
        self.pack(fill=tk.BOTH, expand=True)
        self.create_widgets()
        self.center_window(width, height)
        self.text_entry.focus_set()

    def create_widgets(self) -> None:
        """
        Creates and packs the widgets for the text editor.
        This includes padding frames, labels, and the text entry widget.
        """
        self.left_padding = tk.Frame(self, bg="black", width=20)
        self.left_padding.pack(side="left", fill=tk.Y)

        self.right_padding = tk.Frame(self, bg="black", width=20)
        self.right_padding.pack(side="right", fill=tk.Y)

        self.label_skip_save = tk.Label(
            self, bg="black",
            text="Press 'Ctrl + S' to save or 'Escape' to skip",
            font=("Helvetica", 12), fg="cyan"
        )
        self.label_skip_save.pack(fill=tk.X)

        self.label_limit = tk.Label(
            self, bg="black",
            text=f"Text limit is {self.char_limit} symbols",
            font=("Helvetica", 10), fg="cyan"
        )
        self.label_limit.pack(fill=tk.X)

        custom_font = font.Font(family="Helvetica", size=14, weight="normal")
        self.text_entry = tk.Text(
            self, font=custom_font,
            bg="black", fg="cyan",
            insertbackground="cyan"
        )
        self.text_entry.pack(side="left", fill=tk.BOTH, expand=True)

        self.text_entry.insert(tk.END, self.initial_text)

        self.text_entry.bind("<KeyPress>", self.handle_keypress)

    def handle_keypress(self, event: tk.Event) -> Optional[str]:
        """
        Handles key press events, including enforcing the character limit and
        managing shortcuts.

        Args:
            event (tk.Event): The event object containing information about
            the key press.

        Returns:
            Optional[str]: Returns "break" to prevent further processing of
            the event if the character limit is exceeded.
        """
        current_text: str = self.text_entry.get("1.0", tk.END).strip()

        allowed_keys: list[str] = ['BackSpace', 'Delete',
                                   'Left', 'Right', 'Up',
                                   'Down', 'Home', 'End'
                                   ]
        if len(current_text) >= self.char_limit and \
                event.keysym not in allowed_keys:
            return "break"

        ctrl_key: int = 0x4
        s_key_code: int = 83
        escape_key_code: int = 27
        if (event.state & ctrl_key) and event.keycode == s_key_code:
            self.save_and_close(None)
        elif event.keycode == escape_key_code:
            self.exit_without_saving(None)

        return None

    def save_text(self) -> None:
        """
        Saves the current text from the text entry widget to the `result`
        attribute.
        """
        self.result = self.text_entry.get("1.0", tk.END).strip()

    def save_and_close(self, event: Optional[tk.Event]) -> None:
        """
        Saves the text and closes the application.

        Args:
            event (Optional[tk.Event]): The event object (unused).
        """
        self.save_text()
        self.master.quit()

    def exit_without_saving(self, event: Optional[tk.Event]) -> None:
        """
        Closes the application without saving the text.

        Args:
            event (Optional[tk.Event]): The event object (unused).
        """
        self.master.quit()

    def center_window(self, width: int, height: int) -> None:
        """
        Centers the application window on the screen.

        Args:
            width (int): The width of the window.
            height (int): The height of the window.
        """
        screen_width: int = self.master.winfo_screenwidth()
        screen_height: int = self.master.winfo_screenheight()
        x_coordinate: int = int((screen_width / 2) - (width / 2))
        y_coordinate: int = int((screen_height / 2) - (height / 2))
        self.master.geometry(f"{width}x{height}+{x_coordinate}+{y_coordinate}")



"""
Note module.
"""

from typing import List
from .title import Title
from .text import Text
from .tag import Tag
from .remainder import Reminder
from .created_on import CreatedOn


class Note:
    """
    Class representing a note in the notes book.

    Attributes:
        title (Title): The title of the note.
        text (str): The text of the note.
        tags (List[Tag]): The list of tags in the note.
        created_on (CreatedOn): The created_on of the note.
        reminder (Reminder): The reminder of the note.
    """

    def __init__(self, title: str) -> None:
        """
        Initializes a new Note instance.

        Args:
            title (str): The title of the note.

        Returns:
            None
        """
        self.title: Title = Title(title)
        self.text: Text | None = None
        self.tags: List[Tag] = []
        self.created_on: CreatedOn = CreatedOn()
        self.reminder: Reminder | None = None

    def add_text(self, text: str) -> None:
        """
        Adds text to the note.

        Args:
            text (str): The text to add.

        Returns:
            None

        Raises:
            ValueError: If the text contains lower than 4 symbols and greater
            than 200 symbols.
        """
        normalize_text = text.strip()
        if len(normalize_text) > 200:
            raise ValueError(
                "Text must contains max 200 symbols")
        self.text = Text(normalize_text)

    def add_tags(self, tags: str) -> None:
        """
        Adds tags to the note. Tags are separated by spaces.

        Args:
            tags (str): A string of tags separated by spaces.

        Returns:
            None
        """

        tag_list = tags.split()

        for tag in tag_list:
            if tag.lower() not in self.tags:
                self.tags.append(Tag(tag))

    def remove_tag(self, tag: str):
        """
        Removes a tag from the list of tags in the `Note`
        instance.

        Args:
            tag (str): The tag to remove.

        Raises:
            ValueError: If the tag is not found in the list of tags.

        Returns:
            None
        """
        tag_to_remove = Tag(tag)
        if tag_to_remove in self.tags:
            self.tags.remove(tag)
        else:
            raise ValueError(f"Tag {tag} not found")

    def set_reminder(self, remind_date: str) -> None:
        """
        Sets reminder to the note.

        Args:
            remind_date (str): The remind_date to add.

        Returns:
            None
        """
        self.reminder = Reminder(remind_date)

    def remove_text(self):
        """
        Removes the text from the note.

        Returns:
            None
        """
        self.text = None

    def remove_reminder(self):
        """
        Removes the reminder from the note.

        Returns:
            None
        """
        self.reminder = None

    def __str__(self):
        """
        Returns a string containing the title, text, tags, created_on,
        reminder of note.

        Returns:
        str: A string containing title, text, tags, created_on, reminder of
        note.
        """
        tags_str = ", ".join(map(str, self.tags)) if self.tags else ""
        return (
            f"title: {self.title}\n"
            f"text: {self.text}\n"
            f"tags: {tags_str}\n"
            f"created_on: {self.created_on}\n"
            f"reminder: {self.reminder}"
        )



import difflib


def suggest_command(user_command: str, commands: list[str]) -> list:
    """
    Suggests the closest matching command in case incorrect input.
    """
    closest_matches = difflib.get_close_matches(
        user_command.lower(), commands, n=2, cutoff=0.6
    )

    if closest_matches:
        return closest_matches
    return None



"""
NotesBook module.
"""
from datetime import datetime
from collections import UserDict
from .note import Note


class NotesBook(UserDict):
    """
    Class representing a collection of notes.
    """

    def add_note(self, note: Note) -> None:
        """
        Adds a new note to the collection.

        Args:
            note (Note): The note to be added.

        Raises:
            ValueError: If a note with the same title already exists.
        """
        normalized_title = note.title.value.lower()
        if normalized_title in self.data:
            raise ValueError(f"Note {note.title.value} already exists")
        self.data[normalized_title] = note

    def delete(self, note_title: str) -> None:
        """
        Deletes a note from the collection by its title.

        Args:
            note_title (str): The title of the note to delete.

        Raises:
            ValueError: If the note with the specified title is not found.
        """
        normalized_note_title = note_title.lower()
        if normalized_note_title not in self.data:
            raise ValueError(f"Note {note_title} not found")
        del self.data[normalized_note_title]

    def find(self, note_title: str) -> Note:
        """
        Finds and returns a note by its title.

        Args:
            note_title (str): The title of the note to find.

        Returns:
            Note: The note object with the specified title.

        Raises:
            ValueError: If the note with the specified title is not found.
        """
        normalized_note_title = note_title.lower()
        if normalized_note_title in self.data:
            return self.data[normalized_note_title]
        raise ValueError(f"Note {note_title} not found")

    def search_notes_by_tag(self, tag: str) -> list[Note]:
        """
        Searches for notes containing a specific tag.

        Args:
            tag (str): The tag to search for.

        Returns:
            list[Note]: A list of notes containing the specified tag.
        """
        tag_normalized = tag.strip().lower()
        return [
            note for note in self.data.values()
            if any(t.value in tag_normalized for t in note.tags)
        ]

    def search_notes_by_title(self, title: str) -> list[Note]:
        """
        Searches for notes by title.

        Args:
            title (str): The title to search for.

        Returns:
            list[Note]: A list of notes containing the title.
        """
        title_normalized = title.strip().lower()
        return [
            note for note in self.data.values()
            if title_normalized in note.title.value.lower()
        ]

    def sort_notes_by_tag(self, tag: str) -> list[Note]:
        """
        Sorts notes by the presence of a specific tag.

        Args:
            tag (str): The tag to sort by.

        Returns:
            list[Note]: A list of notes sorted by the presence of the
            specified tag.
        """
        tag_normalized = tag.strip().lower()
        tagged_notes = [note for note in self.data.values() if any(
            t.value.lower() == tag_normalized for t in note.tags)]
        return sorted(tagged_notes, key=lambda note: note.title.value.lower())

    def sort_notes_by_title(self, ascending: bool = True) -> list[Note]:
        """
        Sorts all notes in the collection by their title.

        Args:
            ascending (bool): If True, sorts the notes in ascending order.
                            If False, sorts the notes in descending order.

        Returns:
            list[Note]: A list of notes sorted by their title.
        """
        return sorted(
            self.data.values(),
            key=lambda note: note.title.value.lower(),
            reverse=not ascending
        )

    @property
    def all_tags(self) -> set:
        """
        Returns a set of all unique tags used in the notes.

        Returns:
            set: A set of unique tags.
        """
        return {tag.value for note in self.data.values() for tag in note.tags}

    def upcoming_reminders(self, days: int) -> list:
        """
        Calculate upcoming reminders within a number of days.

        Args:
            days (int): The number of days to check for upcoming reminders.

        Returns:
            list: A list of notes with upcoming reminders within the given
            number of days.
        """
        today = datetime.today().date()
        upcoming_reminders = {}

        for note in self.data.values():
            if note.reminder is None or note.reminder.value < today:
                continue
            days_until_reminder = (note.reminder.value - today).days
            if 0 <= days_until_reminder <= days:
                upcoming_reminders[note.reminder.value] = note
        sorted_reminders = dict(sorted(upcoming_reminders.items()))

        return sorted_reminders.values()



"""
Record module.
"""

from typing import List, Optional
from .name import Name
from .phone import Phone
from .birthday import Birthday
from .address import Address
from .email import Email
from ..constants.validation import validation_errors


class Record:
    """
    Class representing a contact in the address book.

    Attributes:
        name (Name): The name of the contact.
        phones (List[Phone]): The list of phone numbers in the contact.
        birthday (Birthday | None): The birthday of the contact.
        address (Address | None): The address of the contact.
        email (Email | None): The email address of the contact.
    """

    def __init__(self, contact_name: str) -> None:
        """
        Initializes a new Record instance.

        Args:
            contact_name (str): The name of the contact.
        """
        self.name = Name(contact_name)
        self.phones: List[Phone] = []
        self.birthday: Optional[Birthday] = None
        self.email: Optional[Email] = None
        self.address: Optional[Address] = None

    def add_phone(self, phone: str) -> None:
        """
        Adds a new phone number to the record.

        Args:
            phone (str): The phone number to add.

        Raises:
            ValueError: If the phone number already exists in the record.
        """
        new_phone = Phone(phone)
        if new_phone in self.phones:
            raise ValueError(
                validation_errors["duplicate_phone"].format(phone)
            )
        self.phones.append(new_phone)

    def remove_phone(self, phone: str) -> None:
        """
        Removes a phone number from the list of phones in the `Record`
        instance.

        Args:
            phone (str): The phone number to remove.

        Raises:
            ValueError: If the phone number is not found in the list of phones.
        """
        phone_to_remove = Phone(phone)
        if phone_to_remove in self.phones:
            self.phones.remove(phone)
        else:
            raise ValueError(
                validation_errors["phone_not_found"].format(phone)
            )

    def add_birthday(self, birthday: str) -> None:
        """
        Adds a birthday to the record.

        Args:
            birthday (str): The birthday to be added.
        """
        self.birthday = Birthday(birthday)

    def remove_birthday(self) -> None:
        """
        Removes the birthday from the record.
        """
        self.birthday = None

    def add_address(self, address: str) -> None:
        """
        Adds an address to the record.

        Args:
            address (str): The address to be added.
        """
        self.address = Address(address)

    def remove_address(self) -> None:
        """
        Removes the address from the record.

        Returns:
            None
        """
        self.address = None

    def add_email(self, email: str) -> None:
        """
        Adds an email address to the record.

        Args:
            email (str): The email address to be added.
        """
        self.email = Email(email)

    def remove_email(self) -> None:
        """
        Removes the email address from the record.
        """
        self.email = None



"""
Address field module.
"""

from ..constants.validation import (
    ADDRESS_MIN_LENGTH,
    ADDRESS_MAX_LENGTH,
    validation_errors,
)


class Address:
    """
    Class representing an address field.
    """

    def __init__(self, address: str) -> None:
        """
        Initialize an address field.

        Args:
            address (str): The address.

        Raises:
            ValueError: If the address does not meet the validation criteria.
        """
        if not ADDRESS_MIN_LENGTH <= len(address) <= ADDRESS_MAX_LENGTH:
            raise ValueError(validation_errors["invalid_address"])
        self.value = address

    def __str__(self) -> str:
        """
        Return the string representation of the address field.

        Returns:
            str: The string representation of the address field.
        """
        return self.value



"""
Reminder module.
"""

from datetime import datetime, timedelta


class Reminder:
    """
    Class representing a reminder for a specific date.

    Attributes:
        value (datetime): The datetime object representing the reminder date.
    """

    def __init__(self, remind_date: str) -> None:
        """
        Initializes a new Reminder instance and sets the reminder date.

        Args:
            remind_date (str): The reminder date in the format DD.MM.YYYY.

        Raises:
            ValueError: If the date format is invalid or the date is not in
            the future.
        """
        self.set_reminder(remind_date)

    def set_reminder(self, new_date: str) -> None:
        """
        Sets or updates the reminder date.

        Args:
            new_date (str): The new reminder date in the format DD.MM.YYYY.

        Raises:
            ValueError: If the date format is invalid or the date is not in
            the future.
        """
        try:
            remind_datetime = datetime.strptime(new_date, "%d.%m.%Y").date()
        except ValueError as exc:
            raise ValueError("Invalid date format. Use DD.MM.YYYY") from exc

        if remind_datetime <= datetime.now().date():
            raise ValueError("Date must be greater than today")
        self.value = remind_datetime

    def is_reminder_due(self, days: int) -> bool:
        """
        Checks if the reminder is due within a specified number of days.

        Args:
            days (int): The number of days to check if the reminder is due.

        Returns:
            bool: True if the reminder is due within the specified number of
            days, False otherwise.
        """
        today = datetime.now()
        end_date = today + timedelta(days=days)
        return today <= self.value <= end_date

    def __str__(self) -> str:
        """
        Returns a string representation of the reminder date.

        Returns:
            str: The reminder date as a string in the format DD.MM.YYYY.
        """
        return self.value.strftime("%d.%m.%Y")



"""
Address book module.
"""

from typing import List
from collections import UserDict
from datetime import datetime
from fuzzywuzzy import process
from ..helpers.colors import warning
from ..helpers.display import wrap_text, display_table, highlight_term
from .record import Record
from ..constants.validation import DATE_FORMAT, validation_errors
from ..constants.info_messages import info_messages


class AddressBook(UserDict):
    """
    Address book class which holds records.
    """

    def add_record(self, new_record: Record) -> None:
        """
        Adds a new record to the address book.

        Args:
            new_record (Record): The record to be added.

        Raises:
            ValueError: If a contact with the same name already exists.
        """
        normalized_name = new_record.name.value.lower()
        if normalized_name in self.data:
            raise ValueError(
                validation_errors["duplicate_name"]
                .format(new_record.name.value)
            )
        self.data[normalized_name] = new_record

    def find(self, contact_name: str) -> Record:
        """
        Finds a record in the address book.

        Args:
            contact_name (str): The name of the contact.

        Raises:
            ValueError: If the contact is not found.

        Returns:
            Record: The record with the given name.
        """
        normalized_name = contact_name.strip().lower()
        if normalized_name in self.data:
            return self.data[normalized_name]
        raise ValueError(
            validation_errors["name_not_found"].format(contact_name)
        )

    def delete(self, contact_name: str) -> None:
        """
        Deletes a record from the address book.

        Args:
            contact_name (str): The name of the contact.

        Raises:
            ValueError: If the contact is not found.
        """
        normalized_name = contact_name.lower()
        if normalized_name not in self.data:
            raise ValueError(
                validation_errors["name_not_found"].format(contact_name)
            )
        del self.data[normalized_name]

    def upcoming_birthdays(self, days: int, short: bool = False) -> str:
        """
        Calculate upcoming birthdays within a number of days for a given list
        of users.

        Args:
            days (int): The number of days to check for upcoming birthdays.
            short (bool, optional): If True, the output will be shortened.

        Returns:
            dict: A dictionary with "congratulation_date" keys and
            corresponding "name" values for upcoming birthdays within the
            given number of days.
        """
        today = datetime.today().date()
        upcoming_birthdays = {}

        for contact in self.data.values():
            if contact.birthday is None:
                continue
            birthday = contact.birthday.value
            birthday_this_year = birthday.replace(year=today.year)

            if birthday_this_year < today:
                birthday_this_year = birthday.replace(year=today.year + 1)

            if (birthday_this_year - today).days < days:
                upcoming_birthdays[birthday_this_year] = contact.name.value

        sorted_birthdays = [
            [date.strftime(DATE_FORMAT), name]
            for date, name in sorted(upcoming_birthdays.items())
        ]

        if short:
            return (
                info_messages["upcoming_birthdays"]
                .format(len(sorted_birthdays), days)
                if sorted_birthdays else
                info_messages["no_birthdays"].format(days)
            )
        if not sorted_birthdays:
            return warning(info_messages["no_birthdays"].format(days))

        headers = ["Congratulation Date", "Name"]
        return display_table(headers, sorted_birthdays)

    def search(self, search_term: str) -> list:
        """
        Searches for contacts by name or phone number.
        Args:
            search_term (str): The term to search for in the contact names and
            phone numbers.
        Returns:
            list: A list of `Record` instances that match the search term.
        """
        search_term = search_term.lower()
        results = []

        for record in self.data.values():
            if search_term in record.name.value.lower():
                results.append(record)
                continue

            for phone in record.phones:
                if search_term in str(phone):
                    results.append(record)
                    break

        return results

    def smart_search(self, search_term: str, limit: int = 5) -> list:
        """
        Smart search that finds contacts even with typos and suggests contacts
          as the user types.
        Args:
            search_term (str): The term to search for in the contact names and
              phone numbers.
            limit (int): The maximum number of suggestions to return.
            Returns:
            list: A list of `Record` instances that match the search term with
              fuzzy matching.
        """
        search_term = search_term.lower()
        names = [record.name.value for record in self.data.values()]
        phone_numbers = [
            (str(phone), record.name.value)
            for record in self.data.values()
            for phone in record.phones
        ]

        # Get fuzzy matches for names
        name_matches = process.extract(search_term, names, limit=limit)
        matched_names = [
            match[0] for match in name_matches if match[1] >= 70
        ]  # 70% similarity threshold

        # Get matches for phone numbers
        phone_matches = [
            name for phone, name in phone_numbers if search_term in phone
        ]

        # Combine name and phone matches
        matched_names.extend(phone_matches)
        matched_names = list(set(matched_names))  # Remove duplicates

        results = []
        for name in matched_names:
            record = self.data.get(name.lower())  # Fetch the Record object
            if record:  # Ensure that it's a valid Record object
                results.append(record)

        # Return just the list of Record objects
        return results

    @staticmethod
    def display_contacts(
        contacts: List[Record],
        search_term: str = "",
        field: str = ""
    ) -> str:
        """
        Displays a list of contacts in a formatted table.

        Args:
            contacts (List[Record]): A list of Record objects to be displayed.
            search_term (str, optional): The search term used to filter the
            contacts.
            field (str, optional): The field to be searched.

        Returns:
            str: A string representation of the formatted table or empty
            message.
        """
        headers = ["Name", "Phones", "Birthday", "Address", "Email"]
        table = []
        for contact in contacts:
            name = wrap_text(str(contact.name), width=20)
            if search_term and not field or field == "name":
                name = highlight_term(name, search_term)
            phones = "\n".join(map(str, contact.phones))
            if search_term and not field or field == "phones":
                phones = highlight_term(phones, search_term)
            table.append([
                name if contact.name else "-",
                phones if contact.phones else "-",
                str(contact.birthday) if contact.birthday else "-",
                (wrap_text(str(contact.address), width=40)
                    if contact.address else "-"),
                str(contact.email) if contact.email else "-",
            ])

        return display_table(headers, table)



"""
Serialize and deserialize address book data
"""

import pickle
from time import sleep
from typing import Union
from rich.progress import Progress
from ..notes.notes_book import NotesBook
from ..contacts.address_book import AddressBook


def save_data(data: Union["AddressBook", "NotesBook"], filename: str) -> None:
    """
    Saves the provided data to a file.

    Args:
        data: The data to be saved, can be of any serializable type.
        filename (str): The name of the file where the data will be saved.

    Returns:
        None
    """
    with Progress() as progress:
        task = progress.add_task(
            f"[blue]Saving data to {filename}...", total=100
        )
        while not progress.finished:
            progress.update(task, advance=100)
            sleep(0.8)
    with open(filename, "wb") as f:
        pickle.dump(data, f)


def load_data(
    filename: str,
    default_data: Union["AddressBook", "NotesBook"] = None
) -> Union["AddressBook", "NotesBook"]:
    """
    Loads data from a file.

    Args:
        filename (str): The name of the file from which the data will be
        loaded.
        default_data: The default data to return if the file is not found.

    Returns:
        The loaded data or default_data if the file is not found.
    """
    with Progress() as progress:
        task = progress.add_task(
            f"[blue]Loading data from {filename}...", total=100
        )
        while not progress.finished:
            progress.update(task, advance=100)
            sleep(0.8)
    try:
        with open(filename, "rb") as f:
            return pickle.load(f)
    except FileNotFoundError:
        return default_data



"""
Text field module
"""


class Text:
    """
    Class representing a tag field.
    """

    def __init__(self, value: str) -> None:
        """
        Initialize a text field.

        Args:
            value (str): The text of the note.

        Raises:
            ValueError: If the value is empty.
        """
        if not value.strip():
            raise ValueError("Text cannot be empty")
        if len(value.strip()) > 200:
            raise ValueError("Text must be max 200 symbols")
        self.value = value

    def __str__(self) -> str:
        """
        Return the string representation of the text field.

        Returns:
            str: The string representation of the text field.
        """

        return self.value



"""
Phone field module.
"""

import re
from typing import Union
from ..constants.validation import PHONE_PATTERN, validation_errors


class Phone:
    """
    Class representing a phone field.
    """

    def __init__(self, phone: str) -> None:
        """
        Initialize a phone field.

        Args:
            phone (str): The phone number.

        Raises:
            ValueError: If the phone number does not consist of 10 digits.
        """
        if not re.match(PHONE_PATTERN, phone):
            raise ValueError(validation_errors["invalid_phone"])
        self.value = phone

    def __repr__(self) -> str:
        """
        Return the string representation of the phone field.

        Returns:
            str: The string representation of the phone field.
        """
        return self.value

    def __eq__(self, other: Union[str, "Phone"]) -> bool:
        """
        Checks if two Phone instances are equal.

        Args:
            other (Phone): The Phone instance to compare with.

        Returns:
            bool: True if the Phone instances are equal, False otherwise.
        """
        if isinstance(other, Phone):
            return self.value == other.value
        return self.value == other



"""
Generate data module.
"""

from random import randint, random, choice, choices
from datetime import datetime, timedelta
from faker import Faker


def generate_random_contact() -> dict:
    """
    Generates a dictionary of random contact data.

    Returns:
        dict: A dictionary containing random contact data, including name,
        phones, birthday, email, and address.
    """
    fake = Faker()

    data = {
        "name": choice([fake.name(), fake.first_name()]),
        "phones": (
            ["".join(choices("0123456789", k=10)) for _ in range(randint(0, 4))]
        ),
        "birthday": (
            fake.date_of_birth().strftime("%d.%m.%Y") if random() < 0.7 else ""
        ),
        "email": (fake.email() if random() < 0.7 else ""),
        "address": (
            f"{fake.street_address()}, {fake.city()}, {fake.state()}"
            if random() < 0.7
            else ""
        ),
    }

    return data


def generate_random_note() -> dict:
    """
    Generates a dictionary of random note data.

    Returns:
        dict: A dictionary containing random note data, including title,
        text, tags, and reminder.
    """
    fake = Faker()

    data = {
        "title": " ".join(
            choices(fake.words(), k=randint(2, min(5, len(fake.words()))))
        ),
        "text": (fake.text(max_nb_chars=randint(4, 200)) if random() >= 0.3 else ""),
        "tags": [fake.word() for _ in range(randint(0, 5))],
        "reminder": generate_future_date() if random() >= 0.3 else "",
    }

    return data


def generate_future_date():
    """
    Generates a random future date within a specified range.

    Returns:
        str: A string representing the future date in the format DD.MM.YYYY.
    """
    start_date = datetime.today()
    random_days = randint(1, 30)
    return (start_date + timedelta(days=random_days)).strftime("%d.%m.%Y")



"""
Title field module.
"""


class Title:
    """
    Class representing a title field.
    """

    def __init__(self, value: str) -> None:
        """
        Initialize a title field.

        Args:
            value (str): The title of the note.

        Raises:
            ValueError: If the value is empty.
        """
        if not value.strip():
            raise ValueError("Title cannot be empty")
        if len(value.strip()) < 2:
            raise ValueError("Title must be at least 2 symbols")
        self.value = value

    def __str__(self):
        """
        Return the string representation of the title field.

        Returns:
            str: The string representation of the title field.
        """
        return self.value



"""
Name field module.
"""

from ..constants.validation import (
    NAME_MIN_LENGTH,
    NAME_MAX_LENGTH,
    validation_errors
)


class Name:
    """
    Class representing a name field.
    """

    def __init__(self, contact_name: str) -> None:
        """
        Initialize a name field.

        Args:
            contact_name (str): The name of the contact.

        Raises:
            ValueError: If the name is not valid.
        """
        if not NAME_MIN_LENGTH <= len(contact_name) <= NAME_MAX_LENGTH:
            raise ValueError(validation_errors["invalid_name"])
        self.value = contact_name.strip()

    def __str__(self) -> str:
        """
        Return the string representation of the name field.

        Returns:
            str: The string representation of the name field.
        """
        return self.value



from setuptools import setup, find_packages

setup(
    name='motherbot',  # Название вашего пакета
    version='0.2',  # Версия пакета
    packages=find_packages(),  # Автоматически находит все пакеты в проекте
    include_package_data=True,  # Включает любые файлы, указанные в MANIFEST.in
    install_requires=[
        'certifi',
        'cfgv',
        'charset-normalizer',
        'colorama',
        'distlib',
        'docutils ',
        'Faker',
        'filelock',
        'fuzzywuzzy',
        'identify',
        'idna',
        'importlib_metadata',
        'jaraco.classes',
        'jaraco.context',
        'jaraco.functools',
        'keyring',
        'Levenshtein',
        'markdown-it-py',
        'mdurl',
        'more-itertools',
        'motherbot',
        'nh3',
        'nodeenv',
        'pip',
        'pkginfo',
        'platformdirs',
        'pre-commit',
        'prompt_toolkit',
        'pyfiglet',
        'Pygments',
        'python-dateutil',
        'python-Levenshtein',
        'PyYAML',
        'rapidfuzz',
        'readme_renderer',
        'requests',
        'requests-toolbelt',
        'rfc3986',
        'rich',
        'setuptools',
        'six',
        'tabulate',
        'twine',
        'urllib3',
        'virtualenv',
        'wcwidth',
        'wheel',
        'zipp',
    ],
    entry_points={
        'console_scripts': [
            'motherbot=motherbot.main:main',
        ],
    },
    author='Kateryna Tkachenko',
    author_email='tkachenko.city@gmail.com',
    description='Final project for Python course => MotherBot',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Iryna-Holova/goit-pycore-final',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



"""
Input error decorator module.
"""

from typing import Callable


def input_error(func: Callable) -> Callable:
    """
    Decorator that catches exceptions and returns an appropriate message.

    Args:
        func (function): The function to be decorated.

    Returns:
        function: The decorated function.
    """

    def inner(*args, **kwargs):
        """
        The inner function that catches exceptions and returns an appropriate
        message.

        Args:
            *args: The positional arguments passed to the function.
            **kwargs: The keyword arguments passed to the function.

        Returns:
            str: Error message.
        """
        try:
            return func(*args, **kwargs)
        except ValueError as e:
            return str(e)

    return inner



info_messages = {
    "unknown_command": "Unknown command.",
    "operation_cancelled": "Operation canceled.",
    "contact_added": "Contact added successfully.",
    "contact_not_added": "Failed to add contact.",
    "contact_edited": "Contact edited successfully.",
    "contact_deleted": "Contact deleted successfully.",
    "phone_added": "Phone added.",
    "phone_removed": "Phone removed.",
    "birthday_added": "Birthday added.",
    "birthday_removed": "Birthday removed.",
    "email_added": "Email added.",
    "email_removed": "Email removed.",
    "address_added": "Address added.",
    "address_removed": "Address removed.",
    "no_contacts": "Contacts not found.",
    "upcoming_birthdays": "You have {} upcoming birthday(s) in {} days.",
    "no_birthdays": "There are no upcoming birthdays in the next {} days.",
    "fake_contacts_generated": "Fake contacts generated successfully.",
    "note_added": "Note added successfully.",
    "note_edited": "Note edited successfully.",
    "note_deleted": "Note deleted successfully.",
    "text_added": "Text added.",
    "text_removed": "Text removed.",
    "tags_added": "Tags added.",
    "tag_removed": "Tag removed.",
    "reminder_added": "Reminder added.",
    "reminder_removed": "Reminder removed.",
    "no_notes": "There are no notes in the book.",
    "fake_notes_generated": "Fake notes generated successfully.",
    "reminders": "You have {} upcoming reminder(s) in {} days.",
    "no_reminders": "There are no reminders in the next {} days.",
}



"""
Validation constants module.
"""

NAME_MIN_LENGTH = 1
NAME_MAX_LENGTH = 30
PHONE_LENGTH = 10
PHONE_PATTERN = r"^\d{10}$"
DATE_FORMAT = "%d.%m.%Y"
DATE_STR_FORMAT = "DD.MM.YYYY"
EMAIL_PATTERN = r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$"
ADDRESS_MIN_LENGTH = 5
ADDRESS_MAX_LENGTH = 100

validation_errors = {
    "invalid_name": (
        f"Name must consist of {NAME_MIN_LENGTH} "
        f"to {NAME_MAX_LENGTH} characters."
    ),
    "duplicate_name": "Name {} already exists.",
    "name_not_found": "Name {} not found.",
    "invalid_phone": f"Phone number must consist of {PHONE_LENGTH} digits.",
    "duplicate_phone": "Phone number {} already exists.",
    "phone_not_found": "Phone number {} not found.",
    "invalid_birthday": f"Invalid date format. Use {DATE_STR_FORMAT}.",
    "future_birthday": "Birthday cannot be in the future.",
    "invalid_email": "Invalid email address.",
    "invalid_address": (
        f"Address must consist of {ADDRESS_MIN_LENGTH} "
        f"to {ADDRESS_MAX_LENGTH} characters."
    ),
    "duplicate_title": "Title {} already exists.",
    "invalid_number": "Please enter a valid number.",
}



"""
Main module.
"""
from .helpers.suggest import suggest_command
from .helpers.completer import Prompt
from .helpers.serialize import save_data, load_data
from .helpers.colors import green, danger, red
from .helpers.help import get_help
from .helpers.welcome import print_title
from .notes.notes_book import NotesBook
from .contacts.address_book import AddressBook
from .controllers.notes_controllers import (
    add_note,
    change_note,
    delete_note,
    get_notes,
    reminders,
    fake_notes,
    search_notes
)
from .controllers.contacts_controllers import (
    add_contact,
    change_contact,
    get_contacts,
    birthdays,
    delete_contact,
    fake_contacts,
    search_contacts,
    interactive_search_with_autocomplete,
)

controllers = {
    "add-contact": add_contact,
    "change-contact": change_contact,
    "delete-contact": delete_contact,
    "all-contacts": get_contacts,
    "birthdays": birthdays,
    "fake-contacts": fake_contacts,
    "help": get_help,
    "search-contacts": search_contacts,
    "smart-search": interactive_search_with_autocomplete,
}
notes_controllers = {
    "add-note": add_note,
    "change-note": change_note,
    "delete-note": delete_note,
    "all-notes": get_notes,
    "fake-notes": fake_notes,
    "reminders": reminders,
    "search-notes": search_notes
}


def main():
    """
    The main function that serves as the entry point for the application.
    """
    book = load_data("address_book.pkl", default_data=AddressBook())
    notes_book = load_data("notes_book.pkl", default_data=NotesBook())
    print_title("Welcome to the motherbot!", red)
    print(green(book.upcoming_birthdays(days=7, short=True)))
    commands = list(controllers) + list(notes_controllers) + ["close", "exit"]
    prompt = Prompt()

    while True:
        try:
            command = prompt.prompt(
                "Enter a command: ", commands).strip().lower()
        except KeyboardInterrupt:
            print("Good bye!")
            save_data(book, "address_book.pkl")
            save_data(notes_book, "notes_book.pkl")
            break

        if not command:
            continue

        if command in ["close", "exit"]:
            print("Good bye!")
            save_data(notes_book, "notes_book.pkl")
            save_data(book, "address_book.pkl")
            break

        if command == "hello":
            print(green("How can I help you?"))

        elif command in controllers:
            print(controllers[command](book))

        elif command in notes_controllers:
            print(notes_controllers[command](notes_book))

        else:
            similar_commands = suggest_command(command, commands)
            print(danger("Invalid command."))
            if similar_commands and similar_commands != command:
                print("The most similar commands are")
                for cmd in similar_commands:
                    print(f"\t'{cmd}'")


if __name__ == "__main__":
    main()



commands = {
    "add_contact": "add-contact",
    "change_contact": "change-contact",
    "add_phones": "add-phones",
    "remove_phone": "remove-phone",
    "add_birthday": "add-birthday",
    "remove_birthday": "remove-birthday",
    "edit_birthday": "edit-birthday",
    "add_email": "add-email",
    "remove_email": "remove-email",
    "edit_email": "edit-email",
    "add_address": "add-address",
    "remove_address": "remove-address",
    "edit_address": "edit-address",
    "delete_contact": "delete-contact",
    "all_contacts": "all-contacts",
    "birthdays": "birthdays",
    "fake_contacts": "fake-contacts",
    "help": "help",
    "search_contacts": "search-contacts",
    "smart_search": "smart-search",
    "main_menu": "main-menu",
    "add_tags": "add-tags",
    "remove_tag": "remove-tag",
    "add_text": "add-text",
    "remove_text": "remove-text",
    "edit_text": "edit-text",
    "add_reminder": "add-reminder",
    "remove_reminder": "remove-reminder",
    "edit_reminder": "edit-reminder",
}
